@@ -394,12 +394,12 @@ def generate_schema(function, name: Optional[str] = None, description: Optional[
     # append the heartbeat
     # TODO: don't hard-code
     # TODO: if terminal, don't include this
-    if function.__name__ not in ["send_message"]:
-        schema["parameters"]["properties"]["request_heartbeat"] = {
-            "type": "boolean",
-            "description": "Request an immediate heartbeat after function execution. Set to `True` if you want to send a follow-up message or run a follow-up function.",
-        }
-        schema["parameters"]["required"].append("request_heartbeat")
+    # if function.__name__ not in ["send_message"]:
+    schema["parameters"]["properties"]["request_heartbeat"] = {
+        "type": "boolean",
+        "description": "Request an immediate heartbeat after function execution. Set to `True` if you want to send a follow-up message or run a follow-up function.",
+    }
+    schema["parameters"]["required"].append("request_heartbeat")
 
     return schema
 
